class MTError():
    name = ''
    message = ''

    agent = None

    def __init__(self,agent,message):
        self.agent = agent
        self.message = message
    
    def to_xml(self):
        root_container = ElementTree.Element('MTConnectError')
        root_container.append(self.agent.get_header())
        
        error_container = ElementTree.SubElement(root_container, 'Errors')
        error = ElementTree.SubElement(error_container,'Error')
        error.set('errorCode',self.name)
        error.text = self.message

        return ElementTree.tostring(root_container).decode()
        
class MTInvalidRequest(MTError):
    name = 'MTInvalidRequest'

class MTInvalidRange(MTError):
    name = 'MTInvalidRange'



#general imports
import uuid 
from xml.etree import ElementTree

#import mtcitems
from .standard_list import MTC_DataID_list
from .storage import MTDataEntity

#class helper for MTGeneric to display component tree
def tree_helper(component, level=0):
    output_string = '+'+ ('-'*5)*level + str(component.id)+'\n'
    for item in component.get_sub_items():
        output_string = output_string + '+'+('-'*5)*(level+1)+'+'+str(item.id)+'\n'
        
    component_list = component.get_sub_components()
    
    for comp in component_list:
        output_string = output_string + tree_helper(comp,level+1)

    return output_string

#class helper for MTGeneric to get all subcomponents
def item_helper(component):
    output_list = []
    item_list = component.get_sub_items()
    for item in item_list:
        output_list.append(item)
    
    component_list = component.get_sub_components()
    for component in component_list:
        output_list = output_list + item_helper(component)
    
    return output_list
        

class MTGeneric:
    # ! Use: Generic item for all parts of device. Includes components and data_list
    # ? Data: Holds id,name, and attributes as well as parent componnetn and data_list

    #generic variables
    id = None
    name = None

    #attributes
    attributes = {}

    #parrent component
    parent_component = None

    def __init__(self, id, name, component):
        if(id is None):
            raise ValueError('Missing required id for MTC')
        

        self.id = id
        self.name = name
        self.parent_component = component

        self.attributes = {}
        self.data_list = []

    #
    # Modifiers
    #
    def set_name(self, name):
        self.name = name

    def set_id(self,id):
        self.id = id

    #add attribute to container
    def add_attribute(self, name, value):
        self.attributes[name] = value
    
    def __str__(self):
        return "{} with id {} and name {} with parent {}".format(type(self),self.id,self.name,self.parent_component)
    

class MTGenericContainer(MTGeneric):
    #generic variables
    description = None

    #variables used for storage of sub items
    sub_components = {}
    items = {}

    #raw xml data
    xml_data = None

    def __init__(self,id, name, xml_data, parent_component, description=None):
        super().__init__(id, name, parent_component)

        self.xml_data = xml_data
        self.description = description
        self.sub_components = {}
        self.items = {}
        

    #add subaccount directly to container
    def add_subcomponent(self, Component):
        self.sub_components[Component.id] = Component
    
    #add item directly to conrainter
    def add_item(self, Item):
        self.items[Item.id] = Item
    
    #get list of subcomponents
    def get_sub_components(self):
        return list(self.sub_components.values())
    
    #get list of dataitems
    def get_sub_items(self):
        return list(self.items.values())

    #get all sub items of component
    def get_all_sub_items(self):
        return item_helper(self)

    def display_tree(self):
        return tree_helper(self)
    
    def generate_xml(self):
        return ElementTree.tostring(self.xml_data)

class MTDevice(MTGenericContainer):
    #generic variables
    uuid = None

    #variables used for traversal of sub items
    item_dict = {}
    component_dict = {}
    
    def __init__(self,id, name,xml_data, unique_id=None, description=None):
        super().__init__(id, name, xml_data, None, description)
        
        if(unique_id is None):
            unique_id = str(uuid.uuid4())

        self.uuid = unique_id
        self.item_dict={}
        self.component_dict={}
    
    def add_subcomponent(self, Component):
        super().add_subcomponent(Component)
        self.component_dict[Component.id] = Component

    def add_item(self, Item):
        super().add_item(Item)
        self.item_dict[Item.id] = Item

    #add device to device list for traversal
    def add_sub_item(self, Item):
        self.item_dict[Item.id] = Item

    #add component to device list for traversal
    def add_sub_component(self, Component):
        self.component_dict[Component.id] = Component
    
    def get_sub_item(self):
        return self.item_dict.keys()

class MTComponent(MTGenericContainer):
    #descriptor variables
    type = None

    #parent variables
    device = None
    
    def __init__(self,id, name, type, xml_data, parent_component, device, description=None):
        super().__init__(id,name, xml_data, parent_component, description)
  
        self.type = type
        self.device = device

class MTDataItem(MTGeneric):
    category = None
    type = None

    #parent variable
    device = None

    #Last value thats outside the buffer
    ##used in current 
    last_value = None

    #list of data entity assigned
    data_list = []

    def __init__(self, id, name, type, category, device, component):
        

        category = category.upper()
        if(None in [id, type,category, component, device]):
            raise ValueError('Missing required value for DataItem')
        
        if(category not in ['SAMPLE','CONDITION','EVENT']):
            raise ValueError('{} is not a valid category'.format(category))

        if(type not in MTC_DataID_list):
            raise ValueError('{} is not a valid type'.format(type))

        if(id in device.get_sub_item()):
            raise ValueError('id must be unique. {} has been used'.format(id))
        
        super().__init__(id,name,component)

        self.type = type
        self.category = category

        self.device = device

        

    
    #Handle DataEntitys assigned
    def push_data(self, DataEntity):
        if(self.last_value is None):
            self.last_value = DataEntity

        self.data_list.append(DataEntity)
    
    def pop_data(self):
        self.data_list.pop(0)

    def get_data(self, start_sequence=float('-inf'), end_sequence=float('inf')):
        output_list = []
        for data in self.data_list:
            if(data.sequence_number>start_sequence and data.sequence_number<=end_sequence):
                output_list.append(data)
        return output_list
    
    def get_current(self, seq=float('inf')):
        data = self.get_data(end_sequence=seq)
        if(not data):
            return self.last_value

        return data[-1]
    



import os #enviormentl variabels
import datetime #get currenttime
import numbers #verrify value is number
from xml.etree import ElementTree #generate xml

#MTConnect Imports
from .helper import type_to_pascal
class MTDataEntity():
    # ! Use: Handle the individual data values
    # ? Data: Holds sequence number, tiemstamp, and data 

    #sequence number in buffer and timestamp
    sequence_number = None
    timestamp = None

    #The MTDataItem that this data refernced
    dataItem = None

    #actual value
    value = None


    def __init__(self, dataItem, value):
        #set values
        self.dataItem = dataItem
        if(dataItem.category =='SAMPLE' and (not isinstance(value, numbers.Number) and value!="UNAVAILABLE")):
            raise ValueError('SAMPLE value must be number')
            
        if(dataItem.category =='EVENT' and not isinstance(value, str)):
            raise ValueError('EVENT value must be a string')

        if(dataItem.category =='CONDITION' and value not in ['UNAVAILABLE', 'NORMAL', 'WARNING', 'FAULT']):
            raise ValueError('CONDITION value must be one of UNAVAILABLE, NORMAL, WARNING, or FAULT')

        self.value = value
        self.timestamp = datetime.datetime.utcnow()

    #
    # Accessor Functions 
    #
    
    #get timestamp as MTCOnnect formated string
    def get_time_str(self):
        return self.timestamp.strftime("%Y-%m-%dT%H:%M:%SZ")

    #get xml data for current/sample
    def get_xml(self):
        element = ElementTree.Element(type_to_pascal(self.dataItem.type))
        element.set('dataItemId',self.dataItem.id)
        element.set('timestamp',self.get_time_str())
        if(self.dataItem.name is not None):
            element.set('name',self.dataItem.name)

        element.set('sequence',str(self.sequence_number))

        element.text = str(self.value)
        return element


    #
    # Mutator Functions 
    #
    def set_sequence(self, number):
        self.sequence_number = number
        return self.sequence_number


class MTBuffer():
    # ! Use: Handle buffer operations for MTConnect agent
    # ? Data: Holds all data about sequencing and previous values

    #buffer variables
    buffer = None
    buffer_size = None

    #keep track of buffer location
    first_sequence = None #oldest piece of data
    last_sequence = None #newest piece of data
    buffer_pos = 0 #keep track of place in buffer



    #
    # Constructor Functions
    #
    def __init__(self, buffer_length=None):
        #get buffer size
        if(buffer_length is None):
            buffer_length = int(os.environ.get("BUFFER_SIZE",16384))

        #initialize buffer    
        self.buffer_size = buffer_length
        self.buffer = [None] * self.buffer_size

    #
    # Accessor Functions
    #
    def get_buffer(self):
        return self.buffer

    def get_data(self, seq, count=None):
        #if requested sequence is to below sequence
        if(seq<self.first_sequence):
            return ([],self.first_sequence)
        
        #if count is none get max count
        if(count == None):
            count = self.buffer_size

        #if count is <= 0 then return nothing
        if(count <= 0):
            return ([],seq)

        #get location in the buffer
        buffer_loc = seq - self.first_sequence

        #get end position and adjust if it goes over
        end_pos = buffer_loc + count
        if(end_pos >= self.buffer_size):
            end_pos = self.buffer_size

        #get next_sequence calculations
        next_sequence = self.buffer[end_pos-1].sequence_number + 1
        return (self.buffer[buffer_loc:end_pos],next_sequence)
    
    def empty(self):
        return self.first_sequence is None

    def size(self):
        return self.buffer_size    

    #
    # Mutator  Functions
    #
    def push(self, DataElement):
        #if DataElement is not the correct type
        if(not isinstance(DataElement,MTDataEntity)):
            raise TypeError("DataElement is not of type MTDataEntity")

        #if sequence number is greater than the buffer size
        if(self.buffer_pos >= self.buffer_size):
            #get last item
            last_item = self.buffer.pop(0)

            #remove last item from refernced Device
            last_item.dataItem.pop_data()
            
            #update last item list
            last_item.last_value = last_item

            #update sequence
            self.first_sequence = self.buffer[0].sequence_number

            #add last item
            self.buffer.append(None)
            self.buffer_pos = self.buffer_size-1
        
        #get new sequence number
        if(self.last_sequence is None):
            sequence_number = 1
        else:
            sequence_number = self.last_sequence + 1

        #set sequence number and add to list
        DataElement.set_sequence(sequence_number)
        self.buffer[self.buffer_pos] = DataElement

        #add data to refernced dataItem
        DataElement.dataItem.push_data(DataElement)

        #update counter and static variables
        self.buffer_pos = self.buffer_pos + 1
        self.last_sequence = sequence_number
        if(self.first_sequence is None):
            self.first_sequence = sequence_number



#general imports
from xml.etree import ElementTree

#MTConnect imports
from .device import MTDevice, MTComponent, MTDataItem




#
# AGENT XML Helpers
#
def process_path(device_xml,path, item_dict, component_dict):  
    xml_list = device_xml.findall(path)
    component_list = []
    for element in xml_list:
        
        id = element.get('id')
        print(id)
        if(id in item_dict):
            component_list.append(item_dict[id])

        if(id in component_dict):
            component_list.append(component_dict[id])
    return component_list
        

#
# DEVICE XML Helpers
#

#function to process all of the dataitems on a component
def process_dataitem(item_list, device, component):
    for item in item_list:
        #get required objects
        id = item.get('id')
        category = item.get('category')
        type = item.get('type')
        name = item.get('name')

    
        new_item = MTDataItem(id,name, type,category,device,component)
        
        for attribute in item.items():
            new_item.add_attribute(attribute[0], attribute[1])
        component.add_item(new_item)
        device.add_sub_item(new_item)

#function to recursively add components
def process_components(component_list, device, parent_component):
    for component in component_list:
        #get component attributes
        name = component.get('name')
        type = component.tag
        id = component.get('id')
        description = component.find('Description')

        #get the text value for the description
        if(description is not None):
            description = description.text

        #create top level component
        new_component = MTComponent(id, name, type, component, parent_component, device, description)
        device.add_sub_component(new_component)

        parent_component.add_subcomponent(new_component)

        #get list of attributes
        for attribute in component.items():
            new_component.add_attribute(attribute[0], attribute[1])

        #get list of data items
        component_items = component.find('DataItems')
        if(component_items is not None):
            process_dataitem(component_items,device, new_component)

        #get list of subcomponents
        sub_component_item = component.find('Components')
        if(sub_component_item is not None):
            sub_component_list = sub_component_item.getchildren()
            process_components(sub_component_list, device, new_component)        

#read device xml from file
def read_devices(file):
    #read data file
    try:
        device_tree = ElementTree.parse(file)
    except FileNotFoundError:
        raise ValueError('{} is not a valid file'.format(file))

    #list of devices
    device_list = {}

    #get devices
    root = device_tree.getroot()
    for device in root.getchildren():
        #get identifiers for device
        device_name = device.get('name')
        device_uuid = device.get('uuid')
        device_id = device.get('id')
        device_description = device.find('Description')

        #get the text value for the description
        if(device_description is not None):
            device_description = device_description.text

        #create device
        new_device = MTDevice(device_id,device_name,device, device_uuid,device_description)
        
        #get list of attributes
        for attribute in device.items():
            new_device.add_attribute(attribute[0], attribute[1])

        #get list of data items
        device_items = device.find('DataItems')
        if(device_items is not None):
            process_dataitem(device_items,new_device, new_device)

        #get list of subcomponents
        component_item = device.find('Components')
        if(component_item is not None):
            component_list = component_item.getchildren()
            process_components(component_list, new_device, new_device)

        device_list[new_device.id]=new_device

    return (device_list, device_tree)

        
    



MTC_DataID_list = ['ACTUATOR','CHUCK_INTERLOCK','COMMUNICATIONS','DATA_RANGE','DIRECTION','END_OF_BAR','HARDWARE','INTERFACE_STATE','LOGIC_PROGRAM','MOTION_PROGRAM','SYSTEM','ACCELERATION','ACCUMULATED_TIME','ANGULAR_ACCELERATION','ANGULAR_VELOCITY','AMPERAGE','ALTERNATING','DIRECT','ACTUAL','TARGET','ANGLE','ACTUAL','COMMANDED','AXIS_FEEDRATE','ACTUAL','COMMANDED','JOG','PROGRAMMED','RAPID','OVERRIDE','CLOCK_TIME','CONCENTRATION','CONDUCTIVITY','DISPLACEMENT','ELECTRICAL_ENERGY','EQUIPMENT_TIMER','LOADED','WORKING','OPERATING','POWERED','DELAY','FILL_LEVEL','FLOW','FREQUENCY','LENGTH','STANDARD','REMAINING','USEABLE','LINEAR_FORCE','LOAD','MASS','PATH_FEEDRATE','ACTUAL','COMMANDED','JOG','PROGRAMMED','RAPID','OVERRIDE','PATH_POSITION','ACTUAL','COMMANDED','TARGET','PROBE','PH','POSITION','ACTUAL','COMMANDED','PROGRAMMED','TARGET','POWER_FACTOR','PRESSURE','PROCESS_TIMER','PROCESS','DELAY','RESISTANCE','ROTARY_VELOCITY','ACTUAL','COMMANDED','PROGRAMMED','OVERRIDE','SOUND_LEVEL','NO_SCALE','A_SCALE','B_SCALE','C_SCALE','D_SCALE','SPINDLE_SPEED','ACTUAL','COMMANDED','OVERRIDE','STRAIN','TEMPERATURE','TENSION','TILT','TORQUE','VOLT_AMPERE','VOLT_AMPERE_REACTIVE','VELOCITY','VISCOSITY','VOLTAGE','ALTERNATING','DIRECT','ACTUAL','TARGET','WATTAGE','ACTUAL','TARGET','ACTUATOR_STATE','ALARM','ACTIVE_AXES','','','AVAILABILITY','AXIS_COUPLING','','','AXIS_FEEDRATE_OVERRIDE','JOG','PROGRAMMED','RAPID','AXIS_INTERLOCK','AXIS_STATE','BLOCK','BLOCK_COUNT','CHUCK_INTERLOCK','MANUAL_UNCLAMP','CHUCK_STATE','CODE','COMPOSITION_STATE','ACTION','LATERAL','MOTION','SWITCHED','VERTICAL','CONTROLLER_MODE','CONTROLLER_MODE_OVERRIDE','DRY_RUN','SINGLE_BLOCK','MACHINE_AXIS_LOCK','OPTIONAL_STOP','TOOL_CHANGE_STOP','COUPLED_AXES','DIRECTION','ROTARY','LINEAR','DOOR_STATE','END_OF_BAR','PRIMARY','AUXILIARY','EMERGENCY_STOP','EQUIPMENT_MODE','LOADED','WORKING','OPERATING','POWERED','DELAY','EXECUTION','FUNCTIONAL_MODE','HARDNESS','ROCKWELL','VICKERS','SHORE','BRINELL','LEEB','MOHS','INTERFACE_STATE','LINE','MAXIMUM','MINIMUM','LINE_LABEL','LINE_NUMBER','ABSOLUTE','INCREMENTAL','MATERIAL','MESSAGE','OPERATOR_ID','PALLET_ID','PART_COUNT','ALL','GOOD','BAD','TARGET','REMAINING','PART_ID','PART_NUMBER','PATH_FEEDRATE_OVERRIDE','JOG','PROGRAMMED','RAPID','PATH_MODE','POWER_STATE','','','LINE','CONTROL','POWER_STATUS','PROGRAM','PROGRAM_EDIT','PROGRAM_EDIT_NAME','PROGRAM_COMMENT','PROGRAM_HEADER','ROTARY_MODE','ROTARY_VELOCITY_OVERRIDE','SERIAL_NUMBER','SPINDLE_INTERLOCK','TOOL_ASSET_ID','TOOL_NUMBER','TOOL_OFFSET','RADIAL','LENGTH','USER','OPERATOR','MAINTENANCE','SET_UP','WIRE','WORKHOLDING_ID','WORK_OFFSET']


import logging
import os

MTLogger = logging.getLogger(__name__)
level_string = os.getenv('MTCLogLevel','').upper()
level = logging.INFO
if(level_string=='ERROR'):
    level = logging.ERROR
    
elif(level_string=='WARN'):
    level = logging.WARN

elif(level_string=='INFO'):
    level = logging.INFO

elif(level_string=='DEBUG'):
    level = logging.DEBUG

MTLogger.setLevel(level)



#import user test
import unittest

#import storage objects
from mtconnect.storage import MTBuffer, MTDataEntity
from mtconnect.device import MTDataItem, MTDevice

#import libraries for testing
import os
import datetime



class DataEntityTest(unittest.TestCase):
    test_device = None
    test_item_1 =  None

    def __init__(self, *args, **kwargs):
        super(DataEntityTest, self).__init__(*args, **kwargs)

        self.test_device = MTDevice('test_device','1',None)
        self.test_item_1 = MTDataItem('test_1','test_1','SYSTEM','SAMPLE',self.test_device,self.test_device)
    
    def testDataCreation(self):
        entity = MTDataEntity(self.test_item_1,1)

        self.assertTrue(isinstance(entity.timestamp, type(datetime.datetime.now())))
        self.assertEqual(entity.dataItem.id,'test_1')
        self.assertEqual(entity.value,1)

    def testSequenceNumber(self):
        entity = MTDataEntity(self.test_item_1,1)
        entity.set_sequence(1)
        self.assertEqual(entity.sequence_number,1)

class BufferTest(unittest.TestCase):
    test_device = None
    test_item_1 = None
    test_item_2 = None
    
    test_buffer = None

    def __init__(self, *args, **kwargs):
        super(BufferTest, self).__init__(*args, **kwargs)

        self.test_device = MTDevice('test_device','1',None)
        self.test_item_1 = MTDataItem('test_1','test_1','SYSTEM','SAMPLE',self.test_device,self.test_device)
        self.test_item_2 = MTDataItem('test_2','test_2','SYSTEM','SAMPLE',self.test_device,self.test_device)
        
        self.test_buffer = MTBuffer(buffer_length=2)

    def testBufferCreation(self):
        buffer = MTBuffer()
        self.assertEqual(len(buffer.get_buffer()),16384)
        os.environ['BUFFER_SIZE']='10'
        buffer = MTBuffer()
        self.assertEqual(len(buffer.get_buffer()),10)
        buffer = MTBuffer(buffer_length=2)
        self.assertEqual(len(buffer.get_buffer()),2)

    def testBufferPush(self):
        #initalize buffer
        buffer = self.test_buffer

        data_1 = MTDataEntity(self.test_item_1,1)
        data_2 = MTDataEntity(self.test_item_1,2)
        data_3 = MTDataEntity(self.test_item_2,1)

        buffer.push(data_1)
        self.assertEqual(data_1.sequence_number, 1)
        self.assertEqual(buffer.get_buffer(),[data_1,None])
        self.assertEqual(buffer.first_sequence,1)

        buffer.push(data_2)
        self.assertEqual(data_2.sequence_number, 2)
        self.assertEqual(buffer.get_buffer(),[data_1,data_2])
        self.assertEqual(buffer.last_sequence,2)
        self.assertEqual(buffer.first_sequence,1)

        buffer.push(data_3)
        self.assertEqual(data_3.sequence_number, 3)
        self.assertEqual(buffer.get_buffer(),[data_2,data_3])
        self.assertEqual(buffer.last_sequence,3)
        self.assertEqual(buffer.first_sequence,2)
        
    def testBufferGet(self):
        buffer = self.test_buffer

        test_device = MTDevice('test_device','1','1',None)
        test_item = MTDataItem('test_1','test_1','SYSTEM','SAMPLE',test_device,test_device)
        data_1 = MTDataEntity(test_item,1)
        data_2 = MTDataEntity(test_item,2)
        buffer.push(data_1)
        buffer.push(data_2)


        self.assertEqual(buffer.get_data(1,1), ([data_1],2))
        self.assertEqual(buffer.get_data(1,2), ([data_1,data_2],3))
        self.assertEqual(buffer.get_data(1,3), ([data_1,data_2],3))
        



if __name__ == '__main__':
    unittest.main()



#general imports
import os
import uuid  
from xml.etree import ElementTree
from datetime import datetime
from numbers import Number

#mtconnect imports
from .storage import MTBuffer, MTDataEntity
from .xmlhelper import read_devices, process_path
from .error import MTInvalidRequest, MTInvalidRange
from .device import MTComponent, MTDevice
from .loghandler import MTLogger
class MTConnect():
    # ! Use: Handle MTConnect agent
    # ? Data:
    #instanceID 
    instanceId = None
    hostname = None

    #item buffer
    buffer = None
    asset_buffer = None

    #dictionary of devices
    device_dict = None
    device_xml = None

    #used for traversal
    item_dict = {} # list of a
    component_dict = {} # list of components


    def __init__(self,loc='./device.xml',hostname='http://0.0.0.0:80'):
        MTLogger.info("Initializing MTConnect Agent")
        #set variables
        self.hostname = hostname

        #initalize buffer
        self.buffer = MTBuffer()

        #read device information
        file_location = os.getenv('MTCDeviceFile',loc)
        MTLogger.debug('Reading Devices from {}'.format(file_location))
        device_data = read_devices(file_location)

        

        #Update item dict to contain all items
        self.device_dict, self.device_xml  = device_data
        for device in self.device_dict.values():
            self.item_dict.update(device.item_dict)
            self.component_dict.update(device.component_dict)

        #generate instanceId -64bit int uuid4 is 128 so shift it
        self.instanceId = str(uuid.uuid4().int & (1<<64)-1)
        MTLogger.debug('Settings UUID to {}'.format(self.instanceId))

        #create inital values for item
        for item in self.item_dict.values():
            self.push_data(item.id, "UNAVAILABLE")
        
        MTLogger.info("Finished Initalizing MTConnect Agent")

    #
    # Accessor Functions
    #
    def get_device_list(self):
        return self.device_dict.values()
    
    def get_device(self,name=None):
        if(name is None):
            return get_device_list()[0]
        else:
            return self.device_dict[name]
    #
    # Data Functions
    #

    #validate pushing data
    def get_dataId(self,dataId):
        for device in self.device_dict.values():
            if(dataId in device.get_sub_item()):
                return device.item_dict[dataId]
        raise ValueError('DataID {} is not found'.format(dataId))

    #push data from machines
    def push_data(self, dataId, value):
        dataItem = self.get_dataId(dataId)
        new_data = MTDataEntity(dataItem, value)
        self.buffer.push(new_data)


    #
    # MTConnect/XML Functions   
    # 

    def get_header(self):
        header_element = ElementTree.Element('Header')
        header_element.set('version','1.6.0.0')
        header_element.set('instanceId',str(self.instanceId))
        header_element.set('creationTime',datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))
        header_element.set('bufferSize',str(self.buffer.size()))
        header_element.set('assetBufferSize','0')
        header_element.set('assetCount','0')
        return header_element

    #run MTConnect probe command
    def probe(self):
        root_container = ElementTree.Element('MTConnectDevices')
        root_container.append(self.get_header())

        device_container = ElementTree.SubElement(root_container, 'Devices')
        for device in self.device_dict:
            device_container.append(self.device_dict[device].xml_data)
        
        return ElementTree.tostring(root_container).decode()

    #run MTConnect sample command
    def sample(self, path=None, start=None, count=None):
        #data validation

        #if count is given but not start
        if(start is None and count is not None):
            if(count <0):
                start = self.buffer.last_sequence
            else:
                start = self.buffer.first_sequence

        if(start is None):
            start = self.buffer.first_sequence

        if(count is None):
            count = 100

        if(not isinstance(start, Number) and start<0):
            return MTInvalidRequest(self, "Start must be a non negative number").to_xml()

        if(not isinstance(count, Number) ):
            return MTInvalidRequest(self, "Count must be a number").to_xml()
        
        if(self.buffer.empty()):
            return MTInvalidRange(self, "Buffer is currently empty").to_xml()

        if(start < self.buffer.first_sequence or start > self.buffer.last_sequence):
            return MTInvalidRange(self, "Start must be between {} and {}".format(self.buffer.first_sequence, self.buffer.last_sequence)).to_xml()
    
        if(abs(count) > self.buffer.size()):
            return MTInvalidRange(self, "Count must not be greater than {}".format(self.buffer.size()))
        
        #put count and start into usable formats
        if(count < 0):
            start = start + count
            count = abs(count)
        
        end = start + count

        #apply path variable
        if(path is not None):
            component_list = process_path(self.device_xml, path, self.item_dict, self.component_dict)
        else:
            component_list = list(self.device_dict.values())

        #get all sub items from path
        item_list = set()
        for component in component_list:
            item_list = item_list.union(set(component.get_all_sub_items()))
        item_list = list(item_list)

        #get all itesm last dataitem
        current_dict = {}
        for item in item_list:

            #get data 
            data = item.get_data(start, end)
            
            #if device had no data yet then initialize
            if(item.device not in current_dict):
                current_dict[item.device] = {}

            #if component has no data then initialize
            if item.parent_component not in current_dict[item.device]:
                current_dict[item.device][item.parent_component] = {}

            if item.category not in current_dict[item.device][item.parent_component]:
                current_dict[item.device][item.parent_component][item.category]=[]

            current_dict[item.device][item.parent_component][item.category].append(data)

        #format the output xml
        current_stream = ElementTree.Element('Streams')
        for device in current_dict:
            #create root device stream
            device_element = ElementTree.SubElement(current_stream, 'DeviceStream')
            device_element.set('name',device.name)
            device_element.set('uuid',device.uuid)


            #loop through all data iterms and compoments
            for component in current_dict[device]:

                #get root component stream
                stream_element = ElementTree.SubElement(device_element, 'ComponentStream')
                if(isinstance(component, MTComponent)):
                    stream_element.set('component',component.type)
                elif(isinstance(component, MTDevice)):
                    stream_element.set('component','Device')
                stream_element.set('name',component.name)
                stream_element.set('componentId',component.id)

                #get data
                component_data = current_dict[device][component]

                #For each category add data to xml
                for category in component_data:
                    sample_container = ElementTree.SubElement(stream_element, category.title()+'s')
                    for item_list in component_data[category]:
                        for item in item_list:
                            sample_container.append(item.get_xml())
        #format final xml
        root_container = ElementTree.Element('MTConnectStreams')
        root_container.append(self.get_header())
        root_container.append(current_stream)
        return ElementTree.tostring(root_container).decode()


    #run MTConnnect current command
    def current(self, at=None, path=None):
        #data validation
        if(at is None):
            at = self.buffer.first_sequence


        if(not isinstance(at, Number) or at<0):
            return MTInvalidRequest(self, "At must be a non negative number")
        
        if(self.buffer.empty()):
            return MTInvalidRange(self, "Buffer is currently empty")

        if(at < self.buffer.first_sequence or at > self.buffer.last_sequence):
            return MTInvalidRange(self, "At must be between {} and {}".format(self.buffer.first_sequence, self.buffer.last_sequence))

        #apply path variable
        if(path is not None):
            component_list = process_path(self.device_xml, path, self.item_dict, self.component_dict)
        else:
            component_list = list(self.device_dict.values())

        #get all sub items from path
        item_list = set()
        for component in component_list:
            item_list = item_list.union(set(component.get_all_sub_items()))
        item_list = list(item_list)

        #get all itesm last dataitem
        current_dict = {}
        for item in item_list:

            #get data 
            data = item.get_current(at)
            
            #if device had no data yet then initialize
            if(item.device not in current_dict):
                current_dict[item.device] = {}

            #if component has no data then initialize
            if item.parent_component not in current_dict[item.device]:
                current_dict[item.device][item.parent_component] = {}

            if item.category not in current_dict[item.device][item.parent_component]:
                current_dict[item.device][item.parent_component][item.category]=[]
            
            
            
            current_dict[item.device][item.parent_component][item.category].append(data)
        
        #format the output xml
        current_stream = ElementTree.Element('Streams')
        for device in current_dict:
            #create root device stream
            device_element = ElementTree.SubElement(current_stream, 'DeviceStream')
            device_element.set('name',device.name)
            device_element.set('uuid',device.uuid)


            #loop through all data iterms and compoments
            for component in current_dict[device]:

                #get root component stream
                stream_element = ElementTree.SubElement(device_element, 'ComponentStream')
                if(isinstance(component, MTComponent)):
                    stream_element.set('component',component.type)
                elif(isinstance(component, MTDevice)):
                    stream_element.set('component','Device')
                stream_element.set('name',component.name)
                stream_element.set('componentId',component.id)

                #get data
                component_data = current_dict[device][component]

                #For each category add data to xml
                for category in component_data:
                    sample_container = ElementTree.SubElement(stream_element, category.title()+'s')
                    for item in component_data[category]:
                        
                        sample_container.append(item.get_xml())
        #format final xml
        root_container = ElementTree.Element('MTConnectStreams')
        root_container.append(self.get_header())
        root_container.append(current_stream)
        return ElementTree.tostring(root_container).decode()
            


    def error(self, error_text):
        pass



#import user test
import unittest

#import storage objects
from mtconnect.agent import MTConnect

#import libraries for testing
import os
import datetime



class AgentTest(unittest.TestCase):
    test_agent = None

    def __init__(self, *args, **kwargs):
        super(AgentTest, self).__init__(*args, **kwargs)
        self.test_agent = MTConnect(loc='tests/test_device.xml') 
 

    def testProbe(self):
        self.test_agent.probe()
    
    def testCurrent(self):
        self.test_agent.current()
    
    def testSample(self):
        self.test_agent.sample()

if __name__ == '__main__':
    unittest.main()





#helper functions
#convert type to pascal
def type_to_pascal(type):
    words = type.split('_')
    output_string =''
    for word in words:
        if(word in ['PH','AC','DC']):
            output_string=output_string+word
        else:
            output_string=output_string+word.title()
    return output_string